//
// SoundAF.cc
// 
// C++ interface to soundfiles
// based on SGI calls
// dpwe@interval.com 1994aug03
//

#include "SoundAF.h"
#include <assert.h>
#include <stdio.h>
extern "C" {
#include <sndfutil.h>
};

#include "dpwedefs.h"

static const long af_DEFAULT_TRACK = (long)AF_DEFAULT_TRACK;

static const long bufferbytesize = 32768;	// bytes in disk buffer

Sound::Sound(char modechr, double sampRate, int chans, Format fmt)
{   /* Create a new sound object.  Default arguments specified for 
       each parameter in header file (the C++ way). */
    if(modechr != 'r' && modechr != 'w')  {
	// Called illegally
	delete this;
    } else {
	filename = NULL;
	fileHandle = NULL;
	flags = 0;
	pos = 0;
	buffer = NULL;
	bufSize = 0;
	bufConts = 0;
	mode = modechr;
	// F* parameters not set initially; only meaningful when a 
	// file is connected
	setSampleRate(sampRate);
	setChannels(chans);
	setFormat(fmt);
    }
}

Sound::~Sound(void)
{   /* close the soundfile, destroy the sound object */
    if(fileHandle) {
	DBGFPRINTF((stderr, "AFclosefile(%lx)\n", fileHandle));
	AFclosefile(fileHandle);
	fileHandle = NULL;
    }
    if(buffer!=NULL)  {
	delete buffer;
    }
    
}

int SfmtSize(int format)
{   /* return the bytes per sample of a particular data format */
    return (format & 0x0F);	// for now, in lower 4 bits of format word
}

const int BITS_PER_BYTE = 8;

int Sound::open(char *filename)
{   /* open file for input, or create for output.  Returns zero on success. */
    AFfilesetup setup = AFnewfilesetup();
    int rc = 0;
    long sfmt = AF_SAMPFMT_TWOSCOMP, width = BITS_PER_BYTE*SfmtSize(Uformat);

    DBGFPRINTF((stderr, "AFnewsetup %lx\n", setup));

    /* even though it only actually does anything for write, configure 
       the file setup anyway.  man AFopenfile suggests it might be 
       used for interpreting raw data files. */

    DBGFPRINTF((stderr, "AFinit chans %d srate %.0f fmt %d width %d\n", 
		Uchannels, UsampleRate, sfmt, width));
    AFinitfilefmt(setup, AF_FILE_AIFFC);
    AFinitchannels(setup, af_DEFAULT_TRACK, Uchannels);
    AFinitrate(setup, af_DEFAULT_TRACK, UsampleRate);
    AFinitsampfmt(setup, af_DEFAULT_TRACK, sfmt, width);
    if(mode == 'r')  {
	fileHandle = AFopenfile(filename, "r", setup);
	DBGFPRINTF((stderr, "AFopenfileR('%s')=%lx\n", filename, fileHandle));
	if( fileHandle == NULL)  {
	    rc = -1;
	} else {
	    Sound::filename = strdup(filename);
	}
    } else if(mode == 'w') {
	fileHandle = AFopenfile(filename, "w", setup);
	DBGFPRINTF((stderr, "AFopenfileW('%s')=%lx\n", filename, fileHandle));
	if( fileHandle == NULL)  {
	    rc = -1;
	} else {
	    Sound::filename = strdup(filename);
	}
    } else {
	assert("Sound::open mode not r or w"==0);
	rc = -1;
    }
    if(rc == 0)  { 	// opened OK one way or the other
	// Read back the actual file parameters
	FsampleRate = AFgetrate(fileHandle, af_DEFAULT_TRACK);
	Fchannels   = AFgetchannels(fileHandle, af_DEFAULT_TRACK);
	AFgetsampfmt(fileHandle, af_DEFAULT_TRACK, &sfmt, &width);
	DBGFPRINTF((stderr, "AFget sr=%.0f chs=%d fmt=%d wdth=%d\n", 
		    FsampleRate, Fchannels, sfmt, width));
	assert(sfmt == AF_SAMPFMT_TWOSCOMP);	// all we know about
	width = (width + BITS_PER_BYTE - 1) / BITS_PER_BYTE;
	if(width == 2)
	    Fformat = Sfmt_SHORT;
	else if(width == 4)
	    Fformat = Sfmt_LONG;
	else 
	    assert("Sound:open unknown data width"==0);
	if(mode == 'r')  {
	    // user format parameters are set to those of file on open-read
	    UsampleRate = FsampleRate;
	    Uchannels = Fchannels;
	    Uformat = Fformat;
	    // length is set up on open
	    numFrames = AFgetframecnt(fileHandle, af_DEFAULT_TRACK);
	    DBGFPRINTF((stderr, "AFgetfrms=%ld\n", numFrames));
	}
	flags = mode;
    }
    AFfreefilesetup(setup);
    DBGFPRINTF((stderr, "AFfreesetup %lx\n", setup));
    return rc;
}

long Sound::checkBuffer(long bytesize)
{   /* ensure that the disk buffer exists and is <bytesize> bytes big
       & return the worst-case number of frames of relevent format 
       that it can hold */
    long worstframebytes = max(Uchannels,Fchannels)
	                    * max(SfmtSize(Uformat),SfmtSize(Fformat));
    long worstframes = bytesize / worstframebytes;

    if(bytesize > bufSize)  {
	if(buffer != NULL)
	    delete buffer;
	buffer = new byte[bytesize];
	bufSize = bytesize;
	bufConts = 0;
    }
    return worstframes;
}    

// private
long Sound::_readFrames(void *buf, long frames)
{   /* just do the file read without format conversion */
    long frs = AFreadframes(fileHandle, af_DEFAULT_TRACK, buf, frames);

    DBGFPRINTF((stderr, "AFread(%lx,%d)=%ld\n", fileHandle, frames, frs));
    pos += frs;
    return frs;
}

long Sound::readFrames(void *buf, long frames)
{   /* read sound data from file */
    long totalgot = 0;

    if(Fchannels == Uchannels && Fformat == Uformat)  {
        totalgot = _readFrames(buf, frames);
    } else {
	long got = -1, remain = frames;
	long bufferframesize = checkBuffer(bufferbytesize);
	void *opbuf = (Fformat==Uformat)?buf:buffer;
	                 // if only channel conversion required, 
	                 // write its result directly into output

	while(got != 0 && remain > 0)  {
	    got = _readFrames(buffer, lmin(bufferframesize, remain));
	    totalgot += got;
	    // convert channels and format
	    if(Fchannels != Uchannels)  {
		if(Fformat == Sfmt_SHORT)
		    ConvertChannels((short *)buffer, (short *)opbuf, 
				    Fchannels, Uchannels, got);
		else if(Fformat == Sfmt_FLOAT)
		    ConvertFlChans((float *)buffer, (float *)opbuf, 
				   Fchannels, Uchannels, got);
		else {
		    assert("Unable to convert channels - bad format" == 0);
		}
	    }
	    if(Fformat != Uformat)  { 
		ConvertSampleBlockFormat((char *)buffer, (char *)buf, 
					 (int)Fformat, (int)Uformat, 
					 got*Uchannels);
	    }
	    ((byte *)buf) += SfmtSize(Uformat)*Uchannels*got;
	    remain -= got;
	}
    }
    return totalgot;
}

long Sound::_writeFrames(void *buf, long frames)
{   /* write sound data to file */
    long frs = AFwriteframes(fileHandle, af_DEFAULT_TRACK, buf, frames);

    DBGFPRINTF((stderr, "AFwrite(%lx,%d)=%ld\n", fileHandle, frames, frs));
    pos += frs;
    if(pos > numFrames)
	numFrames = pos;
    return frs;
}

long Sound::writeFrames(void *buf, long frames)
{   /* read sound data from file */
    long totaldone = 0;

    if(Fchannels == Uchannels && Fformat == Uformat)  {
        totaldone = _writeFrames(buf, frames);
    } else {
	long todo, done = -1, remain = frames;
	long bufferframesize = checkBuffer(bufferbytesize);
	void *ipbuf = (Fformat==Uformat)?buf:buffer;
	                 // if only channel conversion required, 
	                 // take its input directly from input

	while(done != 0 && remain > 0)  {
	    todo = lmin(bufferframesize, remain);
	    // convert channels and format
	    if(Fformat != Uformat)  { 
		ConvertSampleBlockFormat((char *)buf, (char *)buffer, 
			 		 (int)Uformat, (int)Fformat, 
					 todo*Uchannels);
	    }
	    if(Fchannels != Uchannels)  {
		if(Fformat == Sfmt_SHORT)
		    ConvertChannels((short *)ipbuf, (short *)buffer, 
				    Uchannels, Fchannels, todo);
		else if(Fformat == Sfmt_FLOAT)
		    ConvertFlChans((float *)ipbuf, (float *)buffer, 
				   Uchannels, Fchannels, todo);
		else {
		    assert("Unable to convert channels - bad format" == 0);
		}
	    }
	    done = _writeFrames(buffer, todo);
	    totaldone += done;
	    ((byte *)buf) += SfmtSize(Uformat) * Uchannels * done;
	    remain -= done;
	}
    }
    return totaldone;
}

long Sound::frameTell(void)
{   /* frame index of next to read */
    /* Curiously, AF appears not to supply this */
    return pos;
}

long Sound::frameSeek(long frame)
{   /* move file point to abs pos (in sample frames) */
    if(frame != pos)  {
	pos = AFseekframe(fileHandle, af_DEFAULT_TRACK, frame);
	DBGFPRINTF((stderr, "AFseek(%lx,%d)=%ld\n", fileHandle, frame, pos));
    }
    return pos;
}

int Sound::frameFeof(void)
{   /* return nonzero if at EOF */
    return pos == numFrames;
}

