// -------------------------------------------------------------------
// tkSndWvfm.cc
//
// tk widget to display a sound waveform
// dpwe@mit.edu 1994sep14
//

#include <assert.h>

#include "Sound.h"

extern "C" {
/* INT32 is redefined by X11/Xmd.h (on my IRIX405H) which CC doesn't 
   like, so unset it here */
#undef INT32
#include "tcl.h"
#include "tkConfig.h"
#include "tk.h"
#include "audIO.h"

// Hash table defined in tclSound.cc to convert between sound object 
// tcl names and the actual object pointers.
extern Tcl_HashTable soundTable;

}

// Widget data structure

typedef struct {
    // standard fields
    Tk_Window	tkwin;
    Display	*display;
    Tcl_Interp	*interp;
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    XColor	*fgColor;
    XColor	*selColor;
    int 	borderWidth;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		// appearance of whole window
    char 	*scrollCommand;	// tcl to execute to update scrollbar
    // data for the drawing activities
    GC  	gc;
    int 	doubleBuffer;	// do we double buffer?
    int 	updatePending;
    // fields describing the content
    char	*soundName;	// TCL handle to sound object
    Sound	*sound;		// source of data to display
    int 	viewBase;	// first frame to display
    int 	viewLen;	// number of frames to display
    int 	lastViewBase;	// record last values send to scroll command
    int 	lastViewLen;
    int 	selBase;	// first frame in selected region
    int 	selLen;		// number of frames in selection
    double	vscale;		// vertical scaling
    // buffer for the displayed values
    float	*minMaxBuf;	// buffer holds pairs of (min,max) for eac pxl
    int 	minMaxBufSize;	// how many pairs space is currently allocd for
    int		minMaxBase;	// the viewBase for which current bufconts val
    int		minMaxLen;	// the viewLen for which conts are valid
} SndWvfm;

extern "C" {
// static function prototypes

int SndWvfm_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int SndWvfmCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
static int  SndWvfmConfigure(Tcl_Interp *interp, SndWvfm *sw, 
			     int argc, char **argv, int flags);
static void SndWvfmDestroy(ClientData clientData);
static void SndWvfmDisplay(ClientData clientData);
static void SndWvfmEventProc(ClientData clientData, XEvent *eventPtr);
static int  SndWvfmWidgetCmd(ClientData clientData, Tcl_Interp * interp, 
			     int argc, char **argv);
};

// Argv parsing info

static Tk_ConfigSpec sndWvfmConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background", 
     "#cdb79e", Tk_Offset(SndWvfm, bgBorder), 0 }, 
    {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0}, 
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
     "#701040", Tk_Offset(SndWvfm, fgColor), 0 }, 
    {TK_CONFIG_COLOR, "-selColor", "selColor", "SelColor", 
     "#A0A0A0", Tk_Offset(SndWvfm, selColor), 0 }, 
    {TK_CONFIG_SYNONYM, "-fg", "foreground", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(SndWvfm, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(SndWvfm, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(SndWvfm, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(SndWvfm, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"150", Tk_Offset(SndWvfm, height), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(SndWvfm, relief), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(SndWvfm, doubleBuffer), 0},
    {TK_CONFIG_INT, "-viewBase", "viewBase", "ViewBase",
	"0", Tk_Offset(SndWvfm, viewBase), 0},
    {TK_CONFIG_INT, "-viewLen", "viewLen", "ViewLen",
	"100", Tk_Offset(SndWvfm, viewLen), 0},
    {TK_CONFIG_INT, "-selBase", "selBase", "SelBase",
	"0", Tk_Offset(SndWvfm, selBase), 0},
    {TK_CONFIG_INT, "-selLen", "selLen", "SelLen",
	"0", Tk_Offset(SndWvfm, selLen), 0},
    {TK_CONFIG_STRING, "-scrollCommand", "scrollCommand", "ScrollCommand",
	NULL, Tk_Offset(SndWvfm, scrollCommand), 0},
    {TK_CONFIG_DOUBLE, "-vscale", "vscale", "Vscale",
	"1.0", Tk_Offset(SndWvfm, vscale), 0},
    {TK_CONFIG_STRING, "-sound", "sound", "Sound",
	"<none>", Tk_Offset(SndWvfm, soundName), 0},
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

int SndWvfm_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "sndwvfm", SndWvfmCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

int SndWvfmCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv)
{   /* Tcl command to instance a new SndWvfm widget */
    Tk_Window	main = (Tk_Window) clientData; 
    SndWvfm 	*sw; 
    Tk_Window	tkwin;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be\"", 
			 argv[0], " pathName ?options?\"", NULL);
	return TCL_ERROR;
    }
    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], NULL);
    if(tkwin == NULL)  {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "SndWvfm");

    // Allocate and initialize the structure
    sw = (SndWvfm *)ckalloc(sizeof(SndWvfm));
    sw->tkwin = tkwin;
    sw->display = Tk_Display(tkwin);
    sw->interp  = interp;
    sw->bgBorder= NULL;
    sw->fgColor = NULL;
    sw->selColor = NULL;
    sw->gc = None;
    sw->doubleBuffer = 1;
    sw->updatePending = 0;
    sw->scrollCommand = NULL;
    
    sw->soundName = NULL;
    sw->sound = NULL;
    sw->viewBase  = 0;
    sw->viewLen   = 0;
    sw->lastViewBase  = 0;
    sw->lastViewLen   = 0;
    sw->selBase   = 0;
    sw->selLen    = 0;
    sw->vscale= 1.0;

    sw->minMaxBuf = NULL;
    sw->minMaxBufSize = 0;
    sw->minMaxBase = -1;
    sw->minMaxLen  = 0;

    Tk_CreateEventHandler(sw->tkwin, ExposureMask|StructureNotifyMask,
			  SndWvfmEventProc, (ClientData)sw);
    Tcl_CreateCommand(interp, Tk_PathName(sw->tkwin), SndWvfmWidgetCmd, 
		      (ClientData)sw, (void (*)(ClientData))NULL);
    if(SndWvfmConfigure(interp, sw, argc-2, argv+2, 0) != TCL_OK)  {
	Tk_DestroyWindow(sw->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(sw->tkwin);
    return TCL_OK;
}

static void SndWvfmGetWorkArea(SndWvfm *sw, int *px, int *py, int *pw, int *ph)
{   /* return the area that sound waveform is actually drawn in */
    Tk_Window tkwin = sw->tkwin;
    int x, y, w, h;

    *px = sw->padX;
    *pw = Tk_Width(tkwin) - 2*sw->padX;
    *py = sw->padY;
    *ph = Tk_Height(tkwin) - 2*sw->padY;
}

static int SndWvfmWidgetCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv)
{   /* passed commands executed on the widget */
    SndWvfm	*sw = (SndWvfm *)clientData;
    int 	result = TCL_OK;
    int 	length;
    char	c;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " option ?arg arg ...?\"", NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData)sw);
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result = Tk_ConfigureInfo(interp, sw->tkwin, sndWvfmConfigSpecs, 
				      (char *)sw, NULL, 0);
	} else if(argc==3) {
	    result = Tk_ConfigureInfo(interp, sw->tkwin, sndWvfmConfigSpecs, 
				      (char *)sw, argv[2], 0);
	} else {
	    result = SndWvfmConfigure(interp, sw, argc-2, argv+2, 
				      TK_CONFIG_ARGV_ONLY);
	}
    } else if( (c == 'f') && (strncmp(argv[1], "frameOfPixel", length)==0) )  {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " frameOfPixel x_pixel\"", NULL);
	    return TCL_ERROR;
	}
	// Convert the passed pixel into a frame index
	int pixel, frame, xbase, xwidth, dummy;
	SndWvfmGetWorkArea(sw, &xbase, &dummy, &xwidth, &dummy);
	if(Tcl_GetInt(interp, argv[2], &pixel)!=TCL_OK) {
	    return TCL_ERROR;
	}
	frame = sw->viewBase + ((pixel-xbase)*sw->viewLen)/xwidth;
	sprintf(interp->result, "%d", frame);
    } else if( (c == 'r') && (strncmp(argv[1], "rescan", length)==0) )  {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " rescan\"", NULL);
	    return TCL_ERROR;
	}
	// just a command to make sure we redisplay
	// mark the buffer as invalid
	sw->minMaxBase = -1;
	sw->minMaxLen  = 0;
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure or frameOfPixel", (char *) NULL);
	result = TCL_ERROR;
    }
    if (result == TCL_OK && !sw->updatePending) {
	Tk_DoWhenIdle(SndWvfmDisplay, (ClientData) sw);
	sw->updatePending = 1;
    }
    Tk_Release((ClientData)sw);
    return result;
}

static int SndWvfmConfigure(Tcl_Interp *interp, SndWvfm *sw, 
			    int argc, char **argv, int flags)
{   /* handle configure commands */
    int rslt = TCL_OK;

    if(Tk_ConfigureWidget(interp, sw->tkwin, sndWvfmConfigSpecs, 
			  argc, argv, (char *)sw, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    // somehow convert the soundName into a soundPtr
    Tcl_HashEntry *entryPtr;
    entryPtr = Tcl_FindHashEntry(&soundTable, sw->soundName);
    // (will return NULL if no such sound found)
    if(entryPtr) {
	sw->sound = (Sound *)Tcl_GetHashValue(entryPtr);
	// Ensure that the values are OK
	int solen = sw->sound->getFrames();
	if(sw->viewBase < 0)	    sw->viewBase = 0;
	if(sw->viewLen > solen || sw->viewLen<0)    
	                            sw->viewLen  = solen;
	if(sw->viewBase+sw->viewLen>solen)	
	                            sw->viewBase = solen - sw->viewLen;
	// same for selection
	int selend = sw->selBase + sw->selLen;
	if(sw->selBase < 0)	    sw->selBase = 0;
	if(selend > solen || sw->selLen < 0)
	                            selend  = solen;
	sw->selLen = selend - sw->selBase;
    } else {  // not connected to a real sound
	sw->sound = NULL;
	sw->viewLen = sw->selLen = sw->viewBase = sw->selBase = 0;
    }

    // set the colors and GC for the redisplay
    Tk_SetWindowBackground(sw->tkwin, Tk_3DBorderColor(sw->bgBorder)->pixel);
    if ((sw->gc == None) && (sw->doubleBuffer)) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	sw->gc = Tk_GetGC(sw->tkwin,GCFunction|GCGraphicsExposures, &gcValues);
    }
    // register desired window geometry, arrange for redisplay
    Tk_GeometryRequest(sw->tkwin, sw->width, sw->height);
    Tk_SetInternalBorder(sw->tkwin, sw->borderWidth);
    if (!sw->updatePending) {
	Tk_DoWhenIdle(SndWvfmDisplay, (ClientData)sw);
	sw->updatePending = 1;
    }
    // update the scrollbar ?? now, or after config?
    if(sw->scrollCommand && strlen(sw->scrollCommand)>0
       && (sw->viewLen!=sw->lastViewLen || sw->viewBase!=sw->lastViewBase) ) {
	char args[64];
	
	// args to set command are: totalConts viewSize top bottom
	if(sw->sound != NULL) {
	    sprintf(args, "%d %d %d %d", sw->sound->getFrames(), sw->viewLen, 
		    sw->viewBase, 
		    MIN(sw->sound->getFrames(), sw->viewBase+sw->viewLen));
	} else {
	    strcpy(args, "1 1 0 1");
	}
	rslt = Tcl_VarEval(sw->interp, sw->scrollCommand, " ", args, NULL);
	sw->lastViewLen = sw->viewLen;  sw->lastViewBase = sw->viewBase;
    }
    return rslt;
}

static void SndWvfmEventProc(ClientData clientData, XEvent *eventPtr)
{				/* handle events for the SndWvfm widget */
    SndWvfm *sw = (SndWvfm *)clientData;
    
    if(eventPtr->type == Expose)  {
	if(!sw->updatePending)  {
	    Tk_DoWhenIdle(SndWvfmDisplay, (ClientData)sw);
	    sw->updatePending = 1;
	}
    } else if(eventPtr->type == ConfigureNotify)  {
	// adjust self for resize 
	if(!sw->updatePending)  {
	    Tk_DoWhenIdle(SndWvfmDisplay, (ClientData)sw);
	    sw->updatePending = 1;
	}
    } else if(eventPtr->type == DestroyNotify)  {
	Tcl_DeleteCommand(sw->interp, Tk_PathName(sw->tkwin));
	sw->tkwin = NULL;
	if(sw->updatePending)  {
	    Tk_CancelIdleCall(SndWvfmDisplay, (ClientData)sw);
	}
	Tk_EventuallyFree((ClientData)sw, SndWvfmDestroy);
    }
}

static void drawX(Tk_Window tkwin, GC gc, Drawable d=None, int inset=0)
{				/* write an X into a tkwin */
    Display *disp = Tk_Display(tkwin);
    int x, y, w, h;
    
    if(d==None)
	d = Tk_WindowId(tkwin);
    
    x = inset;
    w = Tk_Width(tkwin) - 2*x;
    y = inset;
    h = Tk_Height(tkwin) - 2*y;
    
    XDrawLine(disp, d, gc, x, y, x+w, y+h);
    XDrawLine(disp, d, gc, x, y+h, x+w, y);
}

extern "C" {
#include "vecops.h"
}

static void SndWvfmScanMinMax(SndWvfm *sw)
{   /* fill up our minMaxBuf for the current view situation */
    int x, y, w, h, i;
    int start, end, pts;
    float *buf, *store;
    double min,max;
    Sound *sound = sw->sound;

    SndWvfmGetWorkArea(sw, &x, &y, &w, &h);
    
    // make sure we have a buffer to save it in too
    if(w != sw->minMaxBufSize)  {
	if(sw->minMaxBuf)		ckfree((char *)sw->minMaxBuf);
	sw->minMaxBuf = (float *)ckalloc(sizeof(float)*w*2); // 2 for min,max
	sw->minMaxBufSize = w;
    }
    store = sw->minMaxBuf;
    pts = 1 + (sw->viewLen/w);
    buf = (float *)ckalloc(sizeof(float)*pts);
    
    // Read data out as a single channel of floats
    sound->setFormat(Sfmt_FLOAT);
    sound->setChannels(1);
    
    end = sw->viewBase;
    for(i = 0; i < w; ++i)  {
	start = end;
	end = sw->viewBase + (((i+1) * sw->viewLen)/ w);
	sound->frameSeek(start);
	pts = (end==start)?1:end-start;
	sound->readFrames(buf, pts);
	VMinMax(buf, 1, pts, &min, &max);
	*store++ = min;
	*store++ = max;
    }
    ckfree((char *)buf);
    sw->minMaxBase = sw->viewBase;
    sw->minMaxLen  = sw->viewLen;
/*    fprintf(stderr, "SWScanMM: size %d base %d len %d\n", 
	    sw->minMaxBufSize, sw->minMaxBase, sw->minMaxLen);
 */
}    

static void SndWvfmDrawWvfm(SndWvfm *sw, int xmin, int xmax, Drawable d=None)
{   /* actually draw some waveform data read from a sound */
    Tk_Window 		tkwin = sw->tkwin;
    GC 			gc    = sw->gc;
    Display 		*disp = Tk_Display(tkwin);
    int 		viewBase = sw->viewBase;
    int 		viewLen  = sw->viewLen;
    int 		selBase  = sw->selBase;
    int 		selLen   = sw->selLen;
    XColor 		*fgColor = sw->fgColor;
    XColor 		*selColor= sw->selColor;
    double 		vscale   = sw->vscale;
    int x, y, w, h, y0, yscale, yl, yh, oyl, oyh, ayl, ayh;
    int i;
    int nxs, xssize = 100;
    XSegment *xsgs, *thisSeg;

    float *minMax;
    
    if(d==None)
	d = Tk_WindowId(tkwin);
    
    if(sw->sound == NULL || viewLen == 0)  {
	drawX(tkwin, gc, d, sw->borderWidth+MAX(sw->padX,sw->padY));
	return;
    }

    SndWvfmGetWorkArea(sw, &x, &y, &w, &h);

    if(sw->minMaxBase != viewBase || sw->minMaxLen != viewLen
       || sw->minMaxBufSize != w)  {
	SndWvfmScanMinMax(sw);
    }
    minMax = sw->minMaxBuf;
    
    y0 = y+h/2;
    yscale = -h/2;
    
    if(selLen > 0)  {
	int selMinX, selMaxX;
	// highlight the selected background
	XSetForeground(disp, gc, selColor->pixel);
	selMinX = x + MAX(0, ((selBase-viewBase)*w)/viewLen);
	selMaxX = x + MIN(w, ((selBase-viewBase+selLen)*w)/viewLen);
	if(selMaxX >= selMinX)  {
	    XFillRectangle(disp, d, gc, selMinX+1, y, MAX(1, selMaxX-selMinX),h);
	}
    }
    
    XSetForeground(disp, gc, fgColor->pixel);
    
    xsgs = (XSegment *)ckalloc(xssize * sizeof(XSegment));
    thisSeg = xsgs;
    nxs = 0;
    for(i = 0; i < w; ++i)  {
	yl = int(y0+yscale*vscale* (*minMax++));
	ayl = yl = MAX(y, MIN(y+h-1, yl));
	yh = int(y0+yscale*vscale* (*minMax++));
	ayh = yh = MIN(y+h-1, MAX(y, yh));
	if(i>0)  {
	    ayl = MIN(yl, oyh);
	    ayh = MAX(yh, oyl);
	}
	oyh = yh;	oyl = yl;
//	XDrawLine(disp, d, gc, x+i, ayl, x+i, ayh);
	thisSeg->x1 = x+i;	thisSeg->y1 = ayl;
	thisSeg->x2 = x+i;	thisSeg++->y2 = ayh;
	if(++nxs == xssize)  {
	    XDrawSegments(disp, d, gc, xsgs, nxs);
	    thisSeg = xsgs;
	    nxs = 0;
	}
    }
    if(nxs > 0)  {
	XDrawSegments(disp, d, gc, xsgs, nxs);
    }
    ckfree((char *)xsgs);
}    

static void SndWvfmDisplay(ClientData clientData)
{				/* actually draw the contents of the widget */
    SndWvfm *sw = (SndWvfm *)clientData;
    Tk_Window tkwin = sw->tkwin;
    Pixmap pm = None;
    Drawable d;
    Display *disp = sw->display;
    int 	tw, th, x, y, w, h, inset;
    
    sw->updatePending = 0;
    if(!Tk_IsMapped(tkwin)) {
	return;
    }
    
    tw = Tk_Width(tkwin);
    th = Tk_Height(tkwin);
    
    // Create pixmap for double-buffering if necessary
    if(sw->doubleBuffer)  {
	pm = XCreatePixmap(disp, Tk_WindowId(tkwin), tw, th, 
			   DefaultDepthOfScreen(Tk_Screen(tkwin)));
	d = pm;
    } else {			// draw directly into window
	d = Tk_WindowId(tkwin);
    }
    // redraw background
    Tk_Fill3DRectangle(disp, d, sw->bgBorder, 0, 0, tw, th, sw->borderWidth, 
		       sw->relief);
    // Actually do the drawing
    inset = 2;
    
    //    drawX(tkwin, sw->gc, d, sw->borderWidth + inset);
    SndWvfmDrawWvfm(sw, 0, 0, d);
    
    // copy to screen if double-buffering
    if(sw->doubleBuffer)  {
	XCopyArea(disp, pm, Tk_WindowId(tkwin), sw->gc, 0, 0, 
		  tw, th, 0, 0);
	XFreePixmap(disp, pm);
    }
}

static void SndWvfmDestroy(ClientData clientData)
{				/* destroy one of our objects */
    SndWvfm *sw = (SndWvfm *)clientData;
    
    Tk_FreeOptions(sndWvfmConfigSpecs, (char *)sw, sw->display, 0);
    if(sw->gc != None)  {
	Tk_FreeGC(sw->display, sw->gc);
    }
    // don't destroy sound as we didn't new it
    ckfree((char *)sw);
}


